# coding: utf-8
"""The IPython ZMQ-based parallel computing interface."""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.


import os
import warnings

import zmq

from traitlets.config.configurable import MultipleInstanceError


from ._version import version_info, __version__
from .serialize import *
from .client.asyncresult import *
from .client.client import Client
from .client.remotefunction import *
from .client.view import *
from .controller.dependency import *
from .error import *
from .util import interactive

#-----------------------------------------------------------------------------
# Functions
#-----------------------------------------------------------------------------


def bind_kernel(**kwargs):
    """Bind an Engine's Kernel to be used as a full IPython kernel.
    
    This allows a running Engine to be used simultaneously as a full IPython kernel
    with the QtConsole or other frontends.
    
    This function returns immediately.
    """
    from ipykernel.kernelapp import IPKernelApp
    from ipyparallel.apps.ipengineapp import IPEngineApp
    
    # first check for IPKernelApp, in which case this should be a no-op
    # because there is already a bound kernel
    if IPKernelApp.initialized() and isinstance(IPKernelApp._instance, IPKernelApp):
        return
    
    if IPEngineApp.initialized():
        try:
            app = IPEngineApp.instance()
        except MultipleInstanceError:
            pass
        else:
            return app.bind_kernel(**kwargs)

    raise RuntimeError("bind_kernel be called from an IPEngineApp instance")


def register_joblib_backend(name='ipyparallel', make_default=False):
    """Register the default ipyparallel backend for joblib."""
    from .joblib import register
    return register(name=name, make_default=make_default)


# nbextension installation (requires notebook ≥ 4.2)
def _jupyter_server_extension_paths():
    return [{
        'module': 'ipyparallel.nbextension'
    }]

def _jupyter_nbextension_paths():
    return [{
        'section': 'tree',
        'src': 'nbextension/static',
        'dest': 'ipyparallel',
        'require': 'ipyparallel/main',
    }]




# encoding: utf-8

"""Utilities to enable code objects to be pickled.

Any process that import this module will be able to pickle code objects.  This
includes the func_code attribute of any function.  Once unpickled, new
functions can be built using new.function(code, globals()).  Eventually
we need to automate all of this so that functions themselves can be pickled.

Reference: A. Tremols, P Cogolo, "Python Cookbook," p 302-305
"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import sys
import types
try:
    import copyreg  # Py 3
except ImportError:
    import copy_reg as copyreg  # Py 2

def code_ctor(*args):
    return types.CodeType(*args)

def reduce_code(co):
    args =  [co.co_argcount, co.co_nlocals, co.co_stacksize,
            co.co_flags, co.co_code, co.co_consts, co.co_names,
            co.co_varnames, co.co_filename, co.co_name, co.co_firstlineno,
            co.co_lnotab, co.co_freevars, co.co_cellvars]
    if sys.version_info[0] >= 3:
        args.insert(1, co.co_kwonlyargcount)
    if sys.version_info > (3, 8):
        args.insert(1, co.co_posonlyargcount)
    return code_ctor, tuple(args)

copyreg.pickle(types.CodeType, reduce_code)



# -*- coding: utf-8 -*-
"""test View objects"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import base64
import sys
import platform
import time
from collections import namedtuple
from tempfile import NamedTemporaryFile

import zmq

import pytest

from IPython import get_ipython
from IPython.utils.io import capture_output
from ipython_genutils.py3compat import unicode_type

import ipyparallel  as pmod
from ipyparallel import error
from ipyparallel import AsyncResult, AsyncHubResult, AsyncMapResult
from ipyparallel.util import interactive

from ipyparallel.tests import add_engines

from .clienttest import ClusterTestCase, crash, wait, skip_without

point = namedtuple("point", "x y")

@pytest.mark.usefixtures('ipython')
class TestView(ClusterTestCase):
    
    def setUp(self):
        # On Win XP, wait for resource cleanup, else parallel test group fails
        if platform.system() == "Windows" and platform.win32_ver()[0] == "XP":
            # 1 sec fails. 1.5 sec seems ok. Using 2 sec for margin of safety
            time.sleep(2)
        super(TestView, self).setUp()

    @pytest.mark.xfail
    def test_z_crash_mux(self):
        """test graceful handling of engine death (direct)"""
        # self.add_engines(1)
        eid = self.client.ids[-1]
        ar = self.client[eid].apply_async(crash)
        self.assertRaisesRemote(error.EngineError, ar.get, 10)
        eid = ar.engine_id
        tic = time.time()
        while eid in self.client.ids and time.time()-tic < 5:
            time.sleep(.01)
        self.assertFalse(eid in self.client.ids, "Engine should have died")
    
    def test_push_pull(self):
        """test pushing and pulling"""
        data = dict(a=10, b=1.05, c=list(range(10)), d={'e':(1,2),'f':'hi'})
        t = self.client.ids[-1]
        v = self.client[t]
        push = v.push
        pull = v.pull
        v.block=True
        nengines = len(self.client)
        push({'data':data})
        d = pull('data')
        self.assertEqual(d, data)
        self.client[:].push({'data':data})
        d = self.client[:].pull('data', block=True)
        self.assertEqual(d, nengines*[data])
        ar = push({'data':data}, block=False)
        self.assertTrue(isinstance(ar, AsyncResult))
        r = ar.get()
        ar = self.client[:].pull('data', block=False)
        self.assertTrue(isinstance(ar, AsyncResult))
        r = ar.get()
        self.assertEqual(r, nengines*[data])
        self.client[:].push(dict(a=10,b=20))
        r = self.client[:].pull(('a','b'), block=True)
        self.assertEqual(r, nengines*[[10,20]])
    
    def test_push_pull_function(self):
        "test pushing and pulling functions"
        def testf(x):
            return 2.0*x
        
        t = self.client.ids[-1]
        v = self.client[t]
        v.block=True
        push = v.push
        pull = v.pull
        execute = v.execute
        push({'testf':testf})
        r = pull('testf')
        self.assertEqual(r(1.0), testf(1.0))
        execute('r = testf(10)')
        r = pull('r')
        self.assertEqual(r, testf(10))
        ar = self.client[:].push({'testf':testf}, block=False)
        ar.get()
        ar = self.client[:].pull('testf', block=False)
        rlist = ar.get()
        for r in rlist:
            self.assertEqual(r(1.0), testf(1.0))
        execute("def g(x): return x*x")
        r = pull(('testf','g'))
        self.assertEqual((r[0](10),r[1](10)), (testf(10), 100))
    
    def test_push_function_globals(self):
        """test that pushed functions have access to globals"""
        @interactive
        def geta():
            return a
        # self.add_engines(1)
        v = self.client[-1]
        v.block=True
        v['f'] = geta
        self.assertRaisesRemote(NameError, v.execute, 'b=f()')
        v.execute('a=5')
        v.execute('b=f()')
        self.assertEqual(v['b'], 5)
    
    def test_push_function_defaults(self):
        """test that pushed functions preserve default args"""
        def echo(a=10):
            return a
        v = self.client[-1]
        v.block=True
        v['f'] = echo
        v.execute('b=f()')
        self.assertEqual(v['b'], 10)

    def test_get_result(self):
        """test getting results from the Hub."""
        c = pmod.Client(profile='iptest')
        # self.add_engines(1)
        t = c.ids[-1]
        v = c[t]
        v2 = self.client[t]
        ar = v.apply_async(wait, 1)
        # give the monitor time to notice the message
        time.sleep(.25)
        ahr = v2.get_result(ar.msg_ids[0], owner=False)
        self.assertIsInstance(ahr, AsyncHubResult)
        self.assertEqual(ahr.get(), ar.get())
        ar2 = v2.get_result(ar.msg_ids[0])
        self.assertNotIsInstance(ar2, AsyncHubResult)
        self.assertEqual(ahr.get(), ar2.get())
        c.close()
    
    def test_run_newline(self):
        """test that run appends newline to files"""
        with NamedTemporaryFile('w', delete=False) as f:
            f.write("""def g():
                return 5
                """)
        v = self.client[-1]
        v.run(f.name, block=True)
        self.assertEqual(v.apply_sync(lambda f: f(), pmod.Reference('g')), 5)

    def test_apply_tracked(self):
        """test tracking for apply"""
        # self.add_engines(1)
        t = self.client.ids[-1]
        v = self.client[t]
        v.block=False
        def echo(n=1024*1024, **kwargs):
            with v.temp_flags(**kwargs):
                return v.apply(lambda x: x, 'x'*n)
        ar = echo(1, track=False)
        ar.wait_for_send(5)
        self.assertTrue(isinstance(ar._tracker, zmq.MessageTracker))
        self.assertTrue(ar.sent)
        ar = echo(track=True)
        ar.wait_for_send(5)
        self.assertTrue(isinstance(ar._tracker, zmq.MessageTracker))
        self.assertEqual(ar.sent, ar._tracker.done)
        ar._tracker.wait()
        self.assertTrue(ar.sent)
        
    def test_push_tracked(self):
        t = self.client.ids[-1]
        ns = dict(x='x'*1024*1024)
        v = self.client[t]
        ar = v.push(ns, block=False, track=False)
        self.assertTrue(ar.sent)
        
        ar = v.push(ns, block=False, track=True)
        ar.wait_for_send()
        self.assertEqual(ar.sent, ar._tracker.done)
        self.assertTrue(ar.sent)
        ar.get()
        
    def test_scatter_tracked(self):
        t = self.client.ids
        x='x'*1024*1024
        ar = self.client[t].scatter('x', x, block=False, track=False)
        self.assertTrue(ar.sent)
        
        ar = self.client[t].scatter('x', x, block=False, track=True)
        ar._sent_event.wait()
        self.assertTrue(isinstance(ar._tracker, zmq.MessageTracker))
        self.assertEqual(ar.sent, ar._tracker.done)
        ar.wait_for_send()
        self.assertTrue(ar.sent)
        ar.get()
    
    def test_remote_reference(self):
        v = self.client[-1]
        v['a'] = 123
        ra = pmod.Reference('a')
        b = v.apply_sync(lambda x: x, ra)
        self.assertEqual(b, 123)


    def test_scatter_gather(self):
        view = self.client[:]
        seq1 = list(range(16))
        view.scatter('a', seq1)
        seq2 = view.gather('a', block=True)
        self.assertEqual(seq2, seq1)
        self.assertRaisesRemote(NameError, view.gather, 'asdf', block=True)
    
    @skip_without('numpy')
    def test_scatter_gather_numpy(self):
        import numpy
        from numpy.testing.utils import assert_array_equal
        view = self.client[:]
        a = numpy.arange(64)
        view.scatter('a', a, block=True)
        b = view.gather('a', block=True)
        assert_array_equal(b, a)
    
    def test_scatter_gather_lazy(self):
        """scatter/gather with targets='all'"""
        view = self.client.direct_view(targets='all')
        x = list(range(64))
        view.scatter('x', x)
        gathered = view.gather('x', block=True)
        self.assertEqual(gathered, x)
        

    @skip_without('numpy')
    def test_apply_numpy(self):
        """view.apply(f, ndarray)"""
        import numpy
        from numpy.testing.utils import assert_array_equal
        
        A = numpy.random.random((100,100))
        view = self.client[-1]
        for dt in [ 'int32', 'uint8', 'float32', 'float64' ]:
            B = A.astype(dt)
            C = view.apply_sync(lambda x:x, B)
            assert_array_equal(B,C)
    
    @skip_without('numpy')
    def test_apply_numpy_object_dtype(self):
        """view.apply(f, ndarray) with dtype=object"""
        import numpy
        from numpy.testing.utils import assert_array_equal
        view = self.client[-1]
        
        A = numpy.array([dict(a=5)])
        B = view.apply_sync(lambda x:x, A)
        assert_array_equal(A,B)
        
        A = numpy.array([(0, dict(b=10))], dtype=[('i', int), ('o', object)])
        B = view.apply_sync(lambda x:x, A)
        assert_array_equal(A,B)
    
    @skip_without('numpy')
    def test_push_pull_recarray(self):
        """push/pull recarrays"""
        import numpy
        from numpy.testing.utils import assert_array_equal
        
        view = self.client[-1]
        
        R = numpy.array([
            (1, 'hi', 0.),
            (2**30, 'there', 2.5),
            (-99999, 'world', -12345.6789),
        ], [('n', int), ('s', '|S10'), ('f', float)])
        
        view['RR'] = R
        R2 = view['RR']
        
        r_dtype, r_shape = view.apply_sync(interactive(lambda : (RR.dtype, RR.shape)))
        self.assertEqual(r_dtype, R.dtype)
        self.assertEqual(r_shape, R.shape)
        self.assertEqual(R2.dtype, R.dtype)
        self.assertEqual(R2.shape, R.shape)
        assert_array_equal(R2, R)

    @skip_without('pandas')
    def test_push_pull_timeseries(self):
        """push/pull pandas.Series"""
        import pandas
        
        ts = pandas.Series(list(range(10)))
        
        view = self.client[-1]
        
        view.push(dict(ts=ts), block=True)
        rts = view['ts']
        
        self.assertEqual(type(rts), type(ts))
        self.assertTrue((ts == rts).all())
    
    def test_map(self):
        view = self.client[:]
        def f(x):
            return x**2
        data = list(range(16))
        r = view.map_sync(f, data)
        self.assertEqual(r, list(map(f, data)))
    
    def test_map_empty_sequence(self):
        view = self.client[:]
        r = view.map_sync(lambda x: x, [])
        self.assertEqual(r, [])
    
    def test_map_iterable(self):
        """test map on iterables (direct)"""
        view = self.client[:]
        # 101 is prime, so it won't be evenly distributed
        arr = range(101)
        # ensure it will be an iterator, even in Python 3
        it = iter(arr)
        r = view.map_sync(lambda x: x, it)
        self.assertEqual(r, list(arr))

    @skip_without('numpy')
    def test_map_numpy(self):
        """test map on numpy arrays (direct)"""
        import numpy
        from numpy.testing.utils import assert_array_equal

        view = self.client[:]
        # 101 is prime, so it won't be evenly distributed
        arr = numpy.arange(101)
        r = view.map_sync(lambda x: x, arr)
        assert_array_equal(r, arr)
    
    def test_scatter_gather_nonblocking(self):
        data = list(range(16))
        view = self.client[:]
        view.scatter('a', data, block=False)
        ar = view.gather('a', block=False)
        self.assertEqual(ar.get(), data)
    
    @skip_without('numpy')
    def test_scatter_gather_numpy_nonblocking(self):
        import numpy
        from numpy.testing.utils import assert_array_equal
        a = numpy.arange(64)
        view = self.client[:]
        ar = view.scatter('a', a, block=False)
        self.assertTrue(isinstance(ar, AsyncResult))
        amr = view.gather('a', block=False)
        self.assertTrue(isinstance(amr, AsyncMapResult))
        assert_array_equal(amr.get(), a)

    def test_execute(self):
        view = self.client[:]
        # self.client.debug=True
        execute = view.execute
        ar = execute('c=30', block=False)
        self.assertTrue(isinstance(ar, AsyncResult))
        ar = execute('d=[0,1,2]', block=False)
        self.client.wait(ar, 1)
        self.assertEqual(len(ar.get()), len(self.client))
        for c in view['c']:
            self.assertEqual(c, 30)
    
    def test_abort(self):
        view = self.client[-1]
        ar = view.execute('import time; time.sleep(1)', block=False)
        ar2 = view.apply_async(lambda: 2)
        view.abort(ar2)
        ar3 = view.apply_async(lambda: 3)
        view.abort(ar3.msg_ids)
        self.assertRaises(error.TaskAborted, ar2.get)
        self.assertRaises(error.TaskAborted, ar3.get)
    
    def test_abort_all(self):
        """view.abort() aborts all outstanding tasks"""
        view = self.client[-1]
        ars = [ view.apply_async(time.sleep, 0.25) for i in range(10) ]
        view.abort()
        view.wait(timeout=5)
        for ar in ars[5:]:
            self.assertRaises(error.TaskAborted, ar.get)
    
    def test_temp_flags(self):
        view = self.client[-1]
        view.block=True
        with view.temp_flags(block=False):
            self.assertFalse(view.block)
        self.assertTrue(view.block)
    
    def test_importer(self):
        view = self.client[-1]
        view.clear(block=True)
        with view.importer:
            import re
        
        @interactive
        def findall(pat, s):
            # this globals() step isn't necessary in real code
            # only to prevent a closure in the test
            re = globals()['re']
            return re.findall(pat, s)
        
        self.assertEqual(view.apply_sync(findall, r'\w+', 'hello world'), 'hello world'.split())
    
    def test_unicode_execute(self):
        """test executing unicode strings"""
        v = self.client[-1]
        v.block=True
        if sys.version_info[0] >= 3:
            code="a='é'"
        else:
            code=u"a=u'é'"
        v.execute(code)
        self.assertEqual(v['a'], u'é')
        
    def test_unicode_apply_result(self):
        """test unicode apply results"""
        v = self.client[-1]
        r = v.apply_sync(lambda : u'é')
        self.assertEqual(r, u'é')
    
    def test_unicode_apply_arg(self):
        """test passing unicode arguments to apply"""
        v = self.client[-1]
        
        @interactive
        def check_unicode(a, check):
            assert not isinstance(a, bytes), "%r is bytes, not unicode"%a
            assert isinstance(check, bytes), "%r is not bytes"%check
            assert a.encode('utf8') == check, "%s != %s"%(a,check)
        
        for s in [ u'é', u'ßø®∫',u'asdf' ]:
            try:
                v.apply_sync(check_unicode, s, s.encode('utf8'))
            except error.RemoteError as e:
                if e.ename == 'AssertionError':
                    self.fail(e.evalue)
                else:
                    raise e
    
    def test_map_reference(self):
        """view.map(<Reference>, *seqs) should work"""
        v = self.client[:]
        v.scatter('n', self.client.ids, flatten=True)
        v.execute("f = lambda x,y: x*y")
        rf = pmod.Reference('f')
        nlist = list(range(10))
        mlist = nlist[::-1]
        expected = [ m*n for m,n in zip(mlist, nlist) ]
        result = v.map_sync(rf, mlist, nlist)
        self.assertEqual(result, expected)

    def test_apply_reference(self):
        """view.apply(<Reference>, *args) should work"""
        v = self.client[:]
        v.scatter('n', self.client.ids, flatten=True)
        v.execute("f = lambda x: n*x")
        rf = pmod.Reference('f')
        result = v.apply_sync(rf, 5)
        expected = [ 5*id for id in self.client.ids ]
        self.assertEqual(result, expected)
    
    def test_eval_reference(self):
        v = self.client[self.client.ids[0]]
        v['g'] = list(range(5))
        rg = pmod.Reference('g[0]')
        echo = lambda x:x
        self.assertEqual(v.apply_sync(echo, rg), 0)
    
    def test_reference_nameerror(self):
        v = self.client[self.client.ids[0]]
        r = pmod.Reference('elvis_has_left')
        echo = lambda x:x
        self.assertRaisesRemote(NameError, v.apply_sync, echo, r)

    def test_single_engine_map(self):
        e0 = self.client[self.client.ids[0]]
        r = list(range(5))
        check = [ -1*i for i in r ]
        result = e0.map_sync(lambda x: -1*x, r)
        self.assertEqual(result, check)
    
    def test_len(self):
        """len(view) makes sense"""
        e0 = self.client[self.client.ids[0]]
        self.assertEqual(len(e0), 1)
        v = self.client[:]
        self.assertEqual(len(v), len(self.client.ids))
        v = self.client.direct_view('all')
        self.assertEqual(len(v), len(self.client.ids))
        v = self.client[:2]
        self.assertEqual(len(v), 2)
        v = self.client[:1]
        self.assertEqual(len(v), 1)
        v = self.client.load_balanced_view()
        self.assertEqual(len(v), len(self.client.ids))

    
    # begin execute tests
    
    def test_execute_reply(self):
        e0 = self.client[self.client.ids[0]]
        e0.block = True
        ar = e0.execute("5", silent=False)
        er = ar.get()
        self.assertEqual(str(er), "<ExecuteReply[%i]: 5>" % er.execution_count)
        self.assertEqual(er.execute_result['data']['text/plain'], '5')

    def test_execute_reply_rich(self):
        e0 = self.client[self.client.ids[0]]
        e0.block = True
        e0.execute("from IPython.display import Image, HTML")
        ar = e0.execute("Image(data=b'garbage', format='png', width=10)", silent=False)
        er = ar.get()
        b64data = base64.b64encode(b'garbage').decode('ascii')
        data, metadata = er._repr_png_()
        assert data.strip() == b64data.strip()
        assert metadata == dict(width=10)
        ar = e0.execute("HTML('<b>bold</b>')", silent=False)
        er = ar.get()
        self.assertEqual(er._repr_html_(), "<b>bold</b>")

    def test_execute_reply_stdout(self):
        e0 = self.client[self.client.ids[0]]
        e0.block = True
        ar = e0.execute("print (5)", silent=False)
        er = ar.get()
        self.assertEqual(er.stdout.strip(), '5')
        
    def test_execute_result(self):
        """execute triggers execute_result with silent=False"""
        view = self.client[:]
        ar = view.execute("5", silent=False, block=True)
        
        expected = [{'text/plain' : '5'}] * len(view)
        mimes = [ out['data'] for out in ar.execute_result ]
        self.assertEqual(mimes, expected)
    
    def test_execute_silent(self):
        """execute does not trigger execute_result with silent=True"""
        view = self.client[:]
        ar = view.execute("5", block=True)
        expected = [None] * len(view)
        self.assertEqual(ar.execute_result, expected)
    
    def test_execute_magic(self):
        """execute accepts IPython commands"""
        view = self.client[:]
        view.execute("a = 5")
        ar = view.execute("%whos", block=True)
        # this will raise, if that failed
        ar.get(5)
        for stdout in ar.stdout:
            lines = stdout.splitlines()
            self.assertEqual(lines[0].split(), ['Variable', 'Type', 'Data/Info'])
            found = False
            for line in lines[2:]:
                split = line.split()
                if split == ['a', 'int', '5']:
                    found = True
                    break
            self.assertTrue(found, "whos output wrong: %s" % stdout)
    
    def test_execute_displaypub(self):
        """execute tracks display_pub output"""
        view = self.client[:]
        view.execute("from IPython.core.display import *")
        ar = view.execute("[ display(i) for i in range(5) ]", block=True)
        
        expected = [ {u'text/plain' : unicode_type(j)} for j in range(5) ]
        for outputs in ar.outputs:
            mimes = [ out['data'] for out in outputs ]
            self.assertEqual(mimes, expected)
    
    def test_apply_displaypub(self):
        """apply tracks display_pub output"""
        view = self.client[:]
        view.execute("from IPython.core.display import *")
        
        @interactive
        def publish():
            [ display(i) for i in range(5) ]
        
        ar = view.apply_async(publish)
        ar.get(5)
        assert ar.wait_for_output(5)
        expected = [ {u'text/plain' : unicode_type(j)} for j in range(5) ]
        for outputs in ar.outputs:
            mimes = [ out['data'] for out in outputs ]
            self.assertEqual(mimes, expected)
    
    def test_execute_raises(self):
        """exceptions in execute requests raise appropriately"""
        view = self.client[-1]
        ar = view.execute("1/0")
        self.assertRaisesRemote(ZeroDivisionError, ar.get, 2)
    
    def test_remoteerror_render_exception(self):
        """RemoteErrors get nice tracebacks"""
        view = self.client[-1]
        ar = view.execute("1/0")
        ip = get_ipython()
        ip.user_ns['ar'] = ar
        with capture_output() as io:
            ip.run_cell("ar.get(2)")
        
        self.assertTrue('ZeroDivisionError' in io.stdout, io.stdout)
    
    def test_compositeerror_render_exception(self):
        """CompositeErrors get nice tracebacks"""
        view = self.client[:]
        ar = view.execute("1/0")
        ip = get_ipython()
        ip.user_ns['ar'] = ar

        with capture_output() as io:
            ip.run_cell("ar.get(2)")
        
        count = min(error.CompositeError.tb_limit, len(view))
        
        self.assertEqual(io.stdout.count('ZeroDivisionError'), count * 2, io.stdout)
        self.assertEqual(io.stdout.count('by zero'), count, io.stdout)
        self.assertEqual(io.stdout.count(':execute'), count, io.stdout)
    
    def test_compositeerror_truncate(self):
        """Truncate CompositeErrors with many exceptions"""
        view = self.client[:]
        requests = []
        for i in range(10):
            requests.append(view.execute("1/0"))
        
        ar = self.client.get_result(requests)
        try:
            ar.get()
        except error.CompositeError as _e:
            e = _e
        else:
            self.fail("Should have raised CompositeError")
        
        lines = e.render_traceback()
        with capture_output() as io:
            e.print_traceback()
        
        self.assertTrue("more exceptions" in lines[-1])
        count = e.tb_limit
        
        self.assertEqual(io.stdout.count('ZeroDivisionError'), 2 * count, io.stdout)
        self.assertEqual(io.stdout.count('by zero'), count, io.stdout)
        self.assertEqual(io.stdout.count(':execute'), count, io.stdout)
    
    def test_magic_pylab(self):
        """%pylab works on engines"""
        pytest.importorskip('matplotlib')
        view = self.client[-1]
        ar = view.execute("%pylab inline")
        # at least check if this raised:
        reply = ar.get(5)
        # include imports, in case user config
        ar = view.execute("plot(rand(100))", silent=False)
        reply = ar.get(5)
        assert ar.wait_for_output(5)
        self.assertEqual(len(reply.outputs), 1)
        output = reply.outputs[0]
        self.assertTrue("data" in output)
        data = output['data']
        self.assertTrue("image/png" in data)
    
    def test_func_default_func(self):
        """interactively defined function as apply func default"""
        def foo():
            return 'foo'
        
        def bar(f=foo):
            return f()
        
        view = self.client[-1]
        ar = view.apply_async(bar)
        r = ar.get(10)
        self.assertEqual(r, 'foo')
    
    def test_data_pub_single(self):
        view = self.client[-1]
        ar = view.execute('\n'.join([
            'from ipyparallel.datapub import publish_data',
            'for i in range(5):',
            '  publish_data(dict(i=i))'
        ]), block=False)
        self.assertTrue(isinstance(ar.data, dict))
        ar.get(5)
        assert ar.wait_for_output(5)
        self.assertEqual(ar.data, dict(i=4))

    def test_data_pub(self):
        view = self.client[:]
        ar = view.execute('\n'.join([
            'from ipyparallel.datapub import publish_data',
            'for i in range(5):',
            '  publish_data(dict(i=i))'
        ]), block=False)
        self.assertTrue(all(isinstance(d, dict) for d in ar.data))
        ar.get(5)
        assert ar.wait_for_output(5)
        self.assertEqual(ar.data, [dict(i=4)] * len(ar))
    
    def test_can_list_arg(self):
        """args in lists are canned"""
        view = self.client[-1]
        view['a'] = 128
        rA = pmod.Reference('a')
        ar = view.apply_async(lambda x: x, [rA])
        r = ar.get(5)
        self.assertEqual(r, [128])

    def test_can_dict_arg(self):
        """args in dicts are canned"""
        view = self.client[-1]
        view['a'] = 128
        rA = pmod.Reference('a')
        ar = view.apply_async(lambda x: x, dict(foo=rA))
        r = ar.get(5)
        self.assertEqual(r, dict(foo=128))

    def test_can_list_kwarg(self):
        """kwargs in lists are canned"""
        view = self.client[-1]
        view['a'] = 128
        rA = pmod.Reference('a')
        ar = view.apply_async(lambda x=5: x, x=[rA])
        r = ar.get(5)
        self.assertEqual(r, [128])

    def test_can_dict_kwarg(self):
        """kwargs in dicts are canned"""
        view = self.client[-1]
        view['a'] = 128
        rA = pmod.Reference('a')
        ar = view.apply_async(lambda x=5: x, dict(foo=rA))
        r = ar.get(5)
        self.assertEqual(r, dict(foo=128))
    
    def test_map_ref(self):
        """view.map works with references"""
        view = self.client[:]
        ranks = sorted(self.client.ids)
        view.scatter('rank', ranks, flatten=True)
        rrank = pmod.Reference('rank')
        
        amr = view.map_async(lambda x: x*2, [rrank] * len(view))
        drank = amr.get(5)
        self.assertEqual(drank, [ r*2 for r in ranks ])
        
    def test_nested_getitem_setitem(self):
        """get and set with view['a.b']"""
        view = self.client[-1]
        view.execute('\n'.join([
            'class A(object): pass',
            'a = A()',
            'a.b = 128',
            ]), block=True)
        ra = pmod.Reference('a')

        r = view.apply_sync(lambda x: x.b, ra)
        self.assertEqual(r, 128)
        self.assertEqual(view['a.b'], 128)

        view['a.b'] = 0

        r = view.apply_sync(lambda x: x.b, ra)
        self.assertEqual(r, 0)
        self.assertEqual(view['a.b'], 0)
    
    def test_return_namedtuple(self):
        def namedtuplify(x, y):
            return point(x, y)
        
        view = self.client[-1]
        p = view.apply_sync(namedtuplify, 1, 2)
        self.assertEqual(p.x, 1)
        self.assertEqual(p.y, 2)

    def test_apply_namedtuple(self):
        def echoxy(p):
            return p.y, p.x

        view = self.client[-1]
        tup = view.apply_sync(echoxy, point(1, 2))
        self.assertEqual(tup, (2,1))

    def test_sync_imports(self):
        view = self.client[-1]
        with capture_output() as io:
            with view.sync_imports():
                import IPython
        self.assertIn("IPython", io.stdout)

        @interactive
        def find_ipython():
            return 'IPython' in globals()

        assert view.apply_sync(find_ipython)

    def test_sync_imports_quiet(self):
        view = self.client[-1]
        with capture_output() as io:
            with view.sync_imports(quiet=True):
                import IPython
        self.assertEqual(io.stdout, '')

        @interactive
        def find_ipython():
            return 'IPython' in globals()

        assert view.apply_sync(find_ipython)

    @skip_without('cloudpickle')
    @pytest.mark.xfail(
        reason="cloudpickle doesn't seem to work right now")
    def test_use_cloudpickle(self):
        view = self.client[:]
        view['_a'] = 'engine'
        sys.modules['__main__']._a = 'client'
        @interactive
        def get_a():
            return _a

        # verify initial condition
        a_list = view.apply_sync(get_a)
        self.assertEqual(a_list, ['engine'] * len(view))

        # enable cloudpickle
        view.use_cloudpickle()
        # cloudpickle respects engine values *if defined*
        a_list = view.apply_sync(get_a)
        self.assertEqual(a_list, ['engine'] * len(view))
        # cloudpickle uses client values if engine doesn't override
        view.execute('del _a', block=True)
        a_list = view.apply_sync(get_a)
        self.assertEqual(a_list, ['client'] * len(view))
        # restore pickle, shouldn't resolve
        view.execute('del _a', block=True)
        view.use_pickle()
        self.assertRaisesRemote(NameError, view.apply_sync, get_a)



"""A simple engine that talks to a controller over 0MQ.
it handles registration, etc. and launches a kernel
connected to the Controller's Schedulers.
"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

from __future__ import print_function

import sys
import time
from getpass import getpass

import zmq
from zmq.eventloop import zmqstream

from jupyter_client.localinterfaces import localhost
from traitlets import (
    Instance, Dict, Integer, Type, Float, Unicode, CBytes, Bool,
    default, observe
)
from ipython_genutils.py3compat import cast_bytes

from ipyparallel.controller.heartmonitor import Heart
from ipyparallel.factory import RegistrationFactory
from ipyparallel.util import disambiguate_url, ioloop

from .kernel import IPythonParallelKernel as Kernel
from ipykernel.kernelapp import IPKernelApp

DEFAULT_MPI_INIT = """
from mpi4py import MPI
mpi_rank = MPI.COMM_WORLD.Get_rank()
mpi_size = MPI.COMM_WORLD.Get_size()
"""

class EngineFactory(RegistrationFactory):
    """IPython engine"""

    # configurables:
    out_stream_factory=Type('ipykernel.iostream.OutStream', config=True,
        help="""The OutStream for handling stdout/err.
        Typically 'ipykernel.iostream.OutStream'""")
    display_hook_factory=Type('ipykernel.displayhook.ZMQDisplayHook', config=True,
        help="""The class for handling displayhook.
        Typically 'ipykernel.displayhook.ZMQDisplayHook'""")
    location=Unicode(config=True,
        help="""The location (an IP address) of the controller.  This is
        used for disambiguating URLs, to determine whether
        loopback should be used to connect or the public address.""")
    timeout=Float(5.0, config=True,
        help="""The time (in seconds) to wait for the Controller to respond
        to registration requests before giving up.""")
    max_heartbeat_misses=Integer(50, config=True,
        help="""The maximum number of times a check for the heartbeat ping of a 
        controller can be missed before shutting down the engine.
        
        If set to 0, the check is disabled.""")
    sshserver=Unicode(config=True,
        help="""The SSH server to use for tunneling connections to the Controller.""")
    sshkey=Unicode(config=True,
        help="""The SSH private key file to use when tunneling connections to the Controller.""")
    paramiko=Bool(sys.platform == 'win32', config=True,
        help="""Whether to use paramiko instead of openssh for tunnels.""")
    
    use_mpi = Bool(False, config=True,
        help="""Enable MPI integration.
        
        If set, MPI rank will be requested for my rank,
        and additionally `mpi_init` will be executed in the interactive shell.
        """
    )
    init_mpi = Unicode(DEFAULT_MPI_INIT, config=True,
        help="""Code to execute in the user namespace when initializing MPI"""
    )
    
    @property
    def tunnel_mod(self):
        from zmq.ssh import tunnel
        return tunnel


    # not configurable:
    connection_info = Dict()
    user_ns = Dict()
    id = Integer(None, allow_none=True, config=True,
        help="""Request this engine ID.
        
        If run in MPI, will use the MPI rank.
        Otherwise, let the Hub decide what our rank should be.
        """
    )
    @default('id')
    def _id_default(self):
        if not self.use_mpi:
            return None
        from mpi4py import MPI
        if MPI.COMM_WORLD.size > 1:
            self.log.debug("MPI rank = %i", MPI.COMM_WORLD.rank)
            return MPI.COMM_WORLD.rank
    
    registrar = Instance('zmq.eventloop.zmqstream.ZMQStream', allow_none=True)
    kernel = Instance(Kernel, allow_none=True)
    hb_check_period=Integer()
    
    # States for the heartbeat monitoring
    # Initial values for monitored and pinged must satisfy "monitored > pinged == False" so that 
    # during the first check no "missed" ping is reported. Must be floats for Python 3 compatibility.
    _hb_last_pinged = 0.0
    _hb_last_monitored = 0.0
    _hb_missed_beats = 0
    # The zmq Stream which receives the pings from the Heart
    _hb_listener = None

    bident = CBytes()
    ident = Unicode()

    @observe('ident')
    def _ident_changed(self, change):
        self.bident = cast_bytes(change['new'])
    using_ssh=Bool(False)


    def __init__(self, **kwargs):
        super(EngineFactory, self).__init__(**kwargs)
        self.ident = self.session.session

    def init_connector(self):
        """construct connection function, which handles tunnels."""
        self.using_ssh = bool(self.sshkey or self.sshserver)

        if self.sshkey and not self.sshserver:
            # We are using ssh directly to the controller, tunneling localhost to localhost
            self.sshserver = self.url.split('://')[1].split(':')[0]

        if self.using_ssh:
            if self.tunnel_mod.try_passwordless_ssh(self.sshserver, self.sshkey, self.paramiko):
                password=False
            else:
                password = getpass("SSH Password for %s: "%self.sshserver)
        else:
            password = False

        def connect(s, url):
            url = disambiguate_url(url, self.location)
            if self.using_ssh:
                self.log.debug("Tunneling connection to %s via %s", url, self.sshserver)
                return self.tunnel_mod.tunnel_connection(s, url, self.sshserver,
                            keyfile=self.sshkey, paramiko=self.paramiko,
                            password=password,
                )
            else:
                return s.connect(url)

        def maybe_tunnel(url):
            """like connect, but don't complete the connection (for use by heartbeat)"""
            url = disambiguate_url(url, self.location)
            if self.using_ssh:
                self.log.debug("Tunneling connection to %s via %s", url, self.sshserver)
                url, tunnelobj = self.tunnel_mod.open_tunnel(url, self.sshserver,
                            keyfile=self.sshkey, paramiko=self.paramiko,
                            password=password,
                )
            return str(url)
        return connect, maybe_tunnel

    def register(self):
        """send the registration_request"""

        self.log.info("Registering with controller at %s"%self.url)
        ctx = self.context
        connect,maybe_tunnel = self.init_connector()
        reg = ctx.socket(zmq.DEALER)
        reg.setsockopt(zmq.IDENTITY, self.bident)
        connect(reg, self.url)
        self.registrar = zmqstream.ZMQStream(reg, self.loop)


        content = dict(uuid=self.ident)
        if self.id is not None:
            self.log.info("Requesting id: %i", self.id)
            content['id'] = self.id
        self.registrar.on_recv(lambda msg: self.complete_registration(msg, connect, maybe_tunnel))
        # print (self.session.key)
        
        self.session.send(self.registrar, "registration_request", content=content)

    def _report_ping(self, msg):
        """Callback for when the heartmonitor.Heart receives a ping"""
        #self.log.debug("Received a ping: %s", msg)
        self._hb_last_pinged = time.time()

    def complete_registration(self, msg, connect, maybe_tunnel):
        # print msg
        self.loop.remove_timeout(self._abort_timeout)
        ctx = self.context
        loop = self.loop
        identity = self.bident
        idents,msg = self.session.feed_identities(msg)
        msg = self.session.deserialize(msg)
        content = msg['content']
        info = self.connection_info
        
        def url(key):
            """get zmq url for given channel"""
            return str(info["interface"] + ":%i" % info[key])
        
        if content['status'] == 'ok':
            if self.id is not None and content['id'] != self.id:
                self.log.warning("Did not get the requested id: %i != %i", content['id'], self.id)
            self.id = content['id']

            # launch heartbeat
            # possibly forward hb ports with tunnels
            hb_ping = maybe_tunnel(url('hb_ping'))
            hb_pong = maybe_tunnel(url('hb_pong'))
            
            hb_monitor = None
            if self.max_heartbeat_misses > 0:
                # Add a monitor socket which will record the last time a ping was seen
                mon = self.context.socket(zmq.SUB)
                mport = mon.bind_to_random_port('tcp://%s' % localhost())
                mon.setsockopt(zmq.SUBSCRIBE, b"")
                self._hb_listener = zmqstream.ZMQStream(mon, self.loop)
                self._hb_listener.on_recv(self._report_ping)
            
            
                hb_monitor = "tcp://%s:%i" % (localhost(), mport)

            heart = Heart(hb_ping, hb_pong, hb_monitor , heart_id=identity)
            heart.start()

            # create Shell Connections (MUX, Task, etc.):
            shell_addrs = url('mux'), url('task')

            # Use only one shell stream for mux and tasks
            stream = zmqstream.ZMQStream(ctx.socket(zmq.ROUTER), loop)
            stream.setsockopt(zmq.IDENTITY, identity)
            shell_streams = [stream]
            for addr in shell_addrs:
                connect(stream, addr)

            # control stream:
            control_addr = url('control')
            control_stream = zmqstream.ZMQStream(ctx.socket(zmq.ROUTER), loop)
            control_stream.setsockopt(zmq.IDENTITY, identity)
            connect(control_stream, control_addr)

            # create iopub stream:
            iopub_addr = url('iopub')
            iopub_socket = ctx.socket(zmq.PUB)
            iopub_socket.setsockopt(zmq.IDENTITY, identity)
            connect(iopub_socket, iopub_addr)
            try:
                from ipykernel.iostream import IOPubThread
            except ImportError:
                pass
            else:
                iopub_socket = IOPubThread(iopub_socket)
                iopub_socket.start()

            # disable history:
            self.config.HistoryManager.hist_file = ':memory:'
            
            # Redirect input streams and set a display hook.
            if self.out_stream_factory:
                sys.stdout = self.out_stream_factory(self.session, iopub_socket, u'stdout')
                sys.stdout.topic = cast_bytes('engine.%i.stdout' % self.id)
                sys.stderr = self.out_stream_factory(self.session, iopub_socket, u'stderr')
                sys.stderr.topic = cast_bytes('engine.%i.stderr' % self.id)
            if self.display_hook_factory:
                sys.displayhook = self.display_hook_factory(self.session, iopub_socket)
                sys.displayhook.topic = cast_bytes('engine.%i.execute_result' % self.id)

            self.kernel = Kernel(parent=self, engine_id=self.id, ident=self.ident, session=self.session,
                    control_stream=control_stream, shell_streams=shell_streams, iopub_socket=iopub_socket,
                    loop=loop, user_ns=self.user_ns, log=self.log)
            
            self.kernel.shell.display_pub.topic = cast_bytes('engine.%i.displaypub' % self.id)
            
                
            # periodically check the heartbeat pings of the controller
            # Should be started here and not in "start()" so that the right period can be taken 
            # from the hubs HeartBeatMonitor.period
            if self.max_heartbeat_misses > 0:
                # Use a slightly bigger check period than the hub signal period to not warn unnecessary 
                self.hb_check_period = int(content['hb_period'])+10
                self.log.info("Starting to monitor the heartbeat signal from the hub every %i ms." , self.hb_check_period)
                self._hb_reporter = ioloop.PeriodicCallback(self._hb_monitor, self.hb_check_period)
                self._hb_reporter.start()
            else:
                self.log.info("Monitoring of the heartbeat signal from the hub is not enabled.")

            
            # FIXME: This is a hack until IPKernelApp and IPEngineApp can be fully merged
            app = IPKernelApp(parent=self, shell=self.kernel.shell, kernel=self.kernel, log=self.log)
            if self.use_mpi and self.init_mpi:
                app.exec_lines.insert(0, self.init_mpi)
            app.init_profile_dir()
            app.init_code()
            
            self.kernel.start()
        else:
            self.log.fatal("Registration Failed: %s"%msg)
            raise Exception("Registration Failed: %s"%msg)

        self.log.info("Completed registration with id %i"%self.id)


    def abort(self):
        self.log.fatal("Registration timed out after %.1f seconds"%self.timeout)
        if self.url.startswith('127.'):
            self.log.fatal("""
            If the controller and engines are not on the same machine,
            you will have to instruct the controller to listen on an external IP (in ipcontroller_config.py):
                c.HubFactory.ip='*' # for all interfaces, internal and external
                c.HubFactory.ip='192.168.1.101' # or any interface that the engines can see
            or tunnel connections via ssh.
            """)
        self.session.send(self.registrar, "unregistration_request", content=dict(id=self.id))
        time.sleep(1)
        sys.exit(255)

    def _hb_monitor(self):
        """Callback to monitor the heartbeat from the controller"""
        self._hb_listener.flush()
        if self._hb_last_monitored > self._hb_last_pinged:
            self._hb_missed_beats += 1
            self.log.warn("No heartbeat in the last %s ms (%s time(s) in a row).", self.hb_check_period, self._hb_missed_beats)
        else:
            #self.log.debug("Heartbeat received (after missing %s beats).", self._hb_missed_beats)
            self._hb_missed_beats = 0

        if self._hb_missed_beats >= self.max_heartbeat_misses:
            self.log.fatal("Maximum number of heartbeats misses reached (%s times %s ms), shutting down.",
                           self.max_heartbeat_misses, self.hb_check_period)
            self.session.send(self.registrar, "unregistration_request", content=dict(id=self.id))
            self.loop.stop()

        self._hb_last_monitored = time.time()
            
        
    def start(self):
        loop = self.loop
        def _start():
            self.register()
            self._abort_timeout = loop.add_timeout(loop.time() + self.timeout, self.abort)
        self.loop.add_callback(_start)


